from pysbr.queries.query import Query


class EventsByMatchup(Query):
    """Get events where two participants matched up against one another.

    This query returns up to n previous head-to-head matches between two participants,
    where n is equal to the count argument. It works with teams or individuals.

    Example:
        EventsByMatchup(1548, 1547, 5) returns from the server the last 5 games between
        the Seahawks and the 49ers.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    Args:
        participant_id1: SBR participant id of one of the participants in the matchup.
            It may refer to a team or an individual.
        participant_id2: SBR participant id of the other participant in the matchup. It
            may refer to a team or an individual.
        count: The max number of events to return, counting in reverse chronological
            order from present.
    """

    @Query.typecheck
    def __init__(self, participant_id1: int, participant_id2: int, count: int):
        super().__init__()
        self.name = "lastMatchupsByParticipants"
        self.arg_str = self._get_args("matchup")
        self.args = {
            "partid1": participant_id1,
            "partid2": participant_id2,
            "limit": count,
        }
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"



from collections import OrderedDict
import itertools
from typing import Dict, List, Union

import pysbr.utils as utils
from pysbr.config.config import Config


class Sport(Config):
    """Provides access to sport and league config files.

    These files hold query-relevant information about a given sport.

    This is a base class not meant to be directly instantiated; see subclasses for each
    league and sport.

    Attributes:
        market_names (Dict[int, str]): Map market id to name. Used by Query.list() and
            Query.dataframe() in order to translate from id to name.
        sport_id (int): The sport's id in SBR's database.
        default_market_id (int): The sport's default betting market id, where the id
            number is from SBR.
        consensus_market_ids (List[int]): Available markets for consensus history query.
    """

    def __init__(self, sport_config: Dict):
        super().__init__()

        self._search_translations = utils.load_yaml(
            utils.build_yaml_path("search_dictionary")
        )

        self._sport = self._translate_dict(
            utils.load_yaml(utils.build_yaml_path(sport_config))
        )

        self._market_ids = self._build_market_ids()

        self.market_names = self._build_market_names(self._sport["markets"])
        self.market_periods = self._build_market_periods(self._sport["markets"])
        self.market_types = self._build_market_types(self._sport["markets"])

        self.sport_id = self._sport["sport id"]
        self.default_market_id = self._sport["default market id"]
        self.consensus_market_ids = self._sport["consensus market ids"]

    def _build_market_ids(self) -> Dict[str, int]:
        """Build the dictionary that is used for searching available betting markets.

        Keys are market group information strings concatenated with market information
        strings, values are market ids.

        Example keys with same value:
            'full game point spreads'
            'full-game pointspread'
            'point spread (including ot)'
        """
        market_ids = {}
        m = self._sport["markets"]
        for x in m:
            keys_a = [v.lower() for v in x.values() if isinstance(v, str)]
            for y in x["market types"]:
                id = y["market id"]
                keys_b = [v.lower() for v in y.values() if isinstance(v, str)]
                keys = [
                    " ".join(el).strip()
                    for el in list(itertools.product(keys_a, keys_b))
                ]
                # Add name on its own for markets other than full game (the
                # concatenated version will not make sense).
                keys.append(y["name"].lower())
                market_ids.update({k: id for k in keys})
        return market_ids

    def _build_market_names(self, m: List[Dict]) -> Dict[int, str]:
        """Build the dictionary that is used to translate from market ids to names."""
        markets = {}
        for x in m:
            for y in x["market types"]:
                markets[y["market id"]] = y["name"]
        return markets

    def _build_market_periods(self, m: List[Dict]) -> Dict[int, List[int]]:
        """Build the dictionary that is used to translate from market ids to periods.

        The periods refer to the parts of a game that a market is active for. For
        example, a 1st quarter or 1st period market will have a value of [1,2].
        """
        markets = {}
        for x in m:
            for y in x["market types"]:
                try:
                    markets[y["market id"]] = x["periods"]
                except KeyError:
                    pass
        return markets

    def _build_market_types(self, m):
        market_types = {}
        name_map = {
            "money-line": "moneyline",
            "pointspread": "spread",
            "totals": "total",
        }
        for x in m:
            for y in x["market types"]:
                try:
                    market_types[y["market id"]] = name_map[y["url"]]
                except KeyError:
                    # futures
                    market_types[y["market id"]] = x["url"]
        return market_types

    def market_ids(self, terms: Union[List[Union[int, str]], int, str]) -> List[int]:
        """Given provided search terms, return a list of matching market ids.

        If search term is string, search for matching betting market. If search term is
        int, assume that it is the ID, and insert it into the list to be returned.

        This method is provided as a convenience so that you don't need to
        remember market id numbers. Case is ignored for search terms.

        A search dictionary is utilized so that you can use common abbrevations /
        spelling variants for markets instead of typing them out / worrying about
        format.

        Example search terms, which all point to the same market id:
            '1st half over/under'
            '1st-half totals'
            '1st half o/u'
            'first half totals'
            'first-half ou'
            '1hou'
            '1htot'
            '1h ou'
            '1h tot'
            '1H OU'
            '1H TOT'
            '1HOU'

        Raises:
            TypeError:
                If a provided search term is not an int or str.
            ValueError:
                If a provided search term string cannot be matched with a market.
        """

        def try_translate(term: str) -> str:
            """Attempt to translate a given search term using the search dictionary.

            If there is no translation for the given string, the original value is
            returned.
            """
            words = term.split(" ")
            translated_words = []
            for w in words:
                try:
                    translated_words.append(search_dict[w])
                except KeyError:
                    translated_words.append(w)
            return " ".join(translated_words)

        def try_match_and_append(term: str, ids: List[int]) -> bool:
            """Attempt to match the search term with a market id.

            If match, append the id to the ids list.
            """
            id = self._market_ids.get(term)
            if id is not None:
                ids.append(id)
                return True
            return False

        def split_word(word: str) -> List[str]:
            """Split a word at index 2 and 3, and join parts with space.

            Split only at index 2 and 3 because the abbreviations in the search
            dictionary are never longer than 3 characters.

            Returned list may be empty for invalid inputs.
            """
            splits = []
            for i in range(2, 4):
                try:
                    splits.append(" ".join([word[:i], word[i:]]))
                except IndexError:
                    pass
            return splits

        def match_term(term: Union[int, str], ids: List[int]) -> bool:
            """Try and match a given search term with a market id."""
            if isinstance(term, int):
                ids.append(term)
                return True
            try:
                term = term.lower().strip()
            except AttributeError:
                raise TypeError("Search terms must be ints or strings.")
            translated_term = try_translate(term)
            if try_match_and_append(translated_term, ids):
                return True
            # The term may be abbreviations concatenated together without a space, for
            # example, '1hou'.
            if len(term.split(" ")) == 1:
                splits = split_word(term)
                for t_split in splits:
                    translated_term = try_translate(t_split)
                    if try_match_and_append(translated_term, ids):
                        return True
            return False

        terms = utils.make_list(terms)
        search_dict = self._search_translations
        ids = []
        for term in terms:
            if not match_term(term, ids):
                raise ValueError(f"Could not find market {term}.")

        return list(OrderedDict.fromkeys(ids))

    def sport_config(
        self,
    ) -> Dict[
        str, Union[int, List[Dict[str, Union[str, List[Dict[str, Union[int, str]]]]]]]
    ]:
        """Get the dictionary created from the sport's config file.

        The dict holds values for 'sport id' and 'default market id', as well as a list
        of all markets available on the sport.
        """
        return self._sport

    def search_translations(self) -> Dict[str, str]:
        """Get the dict containing translations for search abbreviatons / spelling variants.

        This is the dict that is used Sport.market_ids().
        """
        return self._search_translations


class League(Sport):
    """Provides access to league config files.

    These files hold query-relevant information about a given league.

    This is a base class not meant to be directly instantiated; see subclasses for each
    league.

    Attributes:
        league_id (int): The league's id in SBR's database.
        league_name (str): The league's name. May differ from the name given by SBR.
        abbr (str): The standard abbreviation for the league. May differ from the
            abbreviation given by SBR.
    """

    def __init__(self, sport_config, league_config):
        super().__init__(sport_config)

        self._league = self._translate_dict(
            utils.load_yaml(utils.build_yaml_path(league_config))
        )
        self.league_id = self._league["league id"]
        self.league_name = self._league["name"]
        self.abbr = self._league["abbreviation"]

    def league_config(
        self,
    ) -> Dict[str, Union[int, str, List[Dict[str, Union[int, str]]]]]:
        """Get the dictionary created from the league's config file.

        The dict holds values for 'league id', 'name', 'abbreviation'. For leagues
        inheriting from TeamSport, it also holds a list of all teams in the league.
        """
        return self._league


class TeamLeague(League):
    """Additional methods for leagues with team information in their config files.

    Class must be initialized with given sport and league config files.

    This class should not be directly instantiated; use league subclasses.

    'id', 'abbreviation', 'sbr abbreviation', 'name', 'nickname', and 'location' are
    keys provided for each team in the league config file.
    """

    def __init__(self, sport_config, league_config):
        super().__init__(sport_config, league_config)

        self._team_ids = self._build_team_ids()

    def _build_team_ids(self) -> Dict[str, Dict[str, Union[int, List[int]]]]:
        """Build team id search dictionary.

        Values are market ids or list of market ids. If list, the search term is
        ambiguous. For example, if NFL, 'New York' is the value for 'location' in the
        config dictionary for 2 teams (Jets and Giants).
        """

        t = self._league["teams"]
        teams = {}
        for k in [
            "abbreviation",
            "sbr abbreviation",
            "name",
            "nickname",
            "location",
            "full name",
        ]:
            teams[k] = {}
            if not k == "full name":
                for x in t:
                    id_ = x["team id"]
                    k2 = x[k].lower()
                    if k2 in teams[k]:
                        v = teams[k][k2]
                        try:
                            v.append(id_)
                        except AttributeError:
                            teams[k][k2] = [v, id_]
                    else:
                        teams[k][k2] = id_
            else:
                for x in t:
                    full_name = " ".join([x["location"].lower(), x["nickname"].lower()])
                    teams[k][full_name] = x["team id"]

        return teams

    def team_ids(self, terms: Union[int, str, List[Union[int, str]]]) -> List[int]:
        """Given provided search terms, return a list of matching team ids.

        If search term is string, search for matching team. If search term is
        int, assume that it is the ID, and insert it into the list to be returned.

        This method is provided as a convenience so that you don't need to
        remember team id numbers. Case is ignored for search terms.

        Example search terms:
            Seattle
            Seahawks
            Seattle Seahawks
            sea
            SEA

        Raises:
            TypeError:
                If a search term is not an int or str.
            ValueError:
                If a search term is ambiguous, or cannot be matched with a team.
        """

        terms = utils.make_list(terms)
        ids = []
        for t in terms:
            if isinstance(t, int):
                ids.append(t)
            else:
                old_t = t
                try:
                    t = t.lower()
                except AttributeError:
                    raise TypeError("Search terms must be ints or strings.")
                try:
                    match = [v[t] for k, v in self._team_ids.items() if t in v][0]
                    ids.append(match)
                except IndexError:
                    raise ValueError(f"Could not find team {old_t}.")
                if isinstance(match, list):
                    raise ValueError(
                        utils.str_format(
                            f"""Search term '{old_t}' is ambiguous.
                            Matching ids: {', '.join([str(m) for m in match])}.
                        """,
                            squish=True,
                        )
                    )
        return list(OrderedDict.fromkeys(ids))


class Football(Sport):
    """Provides access to football config files."""

    def __init__(self):
        super().__init__("football")


class Basketball(Sport):
    """Provides access to basketball config files."""

    def __init__(self):
        super().__init__("basketball")


class Baseball(Sport):
    """Provides access to baseball config files."""

    def __init__(self):
        super().__init__("baseball")


class Hockey(Sport):
    """Provides access to hockey config files."""

    def __init__(self):
        super().__init__("hockey")


class Soccer(Sport):
    """Provides access to soccer config files."""

    def __init__(self):
        super().__init__("soccer")


class Tennis(Sport):
    """Provides access to tennis config files."""

    def __init__(self):
        super().__init__("tennis")


class Fighting(Sport):
    """Provides access to tennis config files."""

    def __init__(self):
        super().__init__("fighting")


class NFL(TeamLeague):
    """Provides access to NFL config files."""

    def __init__(self):
        super().__init__("football", "nfl")


class NCAAF(TeamLeague):
    """Provides access to NCAAF config files."""

    def __init__(self):
        super().__init__("football", "ncaaf")


class MLB(League):
    """Provides access to MLB config files."""

    def __init__(self):
        super().__init__("baseball", "mlb")


class NBA(League):
    """Provides access to NBA config files."""

    def __init__(self):
        super().__init__("basketball", "nba")


class NCAAB(TeamLeague):
    """Provides access to NCAAB config files."""

    def __init__(self):
        super().__init__("basketball", "ncaab")

        self._market_ids = self._build_market_ids()

        self.market_names = self._build_market_names(self._league["markets"])
        self.market_periods = self._build_market_periods(self._league["markets"])
        self.market_types = self._build_market_types(self._league["markets"])


class NHL(League):
    """Provides access to NHL config files."""

    def __init__(self):
        super().__init__("hockey", "nhl")


class EPL(League):
    """Provides access to EPL config files."""

    def __init__(self):
        super().__init__("soccer", "epl")


class UCL(League):
    """Provides access to UCL config files."""

    def __init__(self):
        super().__init__("soccer", "ucl")


class LaLiga(League):
    """Provides access to La Liga config files."""

    def __init__(self):
        super().__init__("soccer", "laliga")


class Bundesliga(League):
    """Provides access to Bundesliga config files."""

    def __init__(self):
        super().__init__("soccer", "bundesliga")


class UEFANationsLeague(League):
    """Provides access to UEFA Nations League config files."""

    def __init__(self):
        super().__init__("soccer", "uefanationsleague")


class ATP(League):
    """Provides access to ATP config files."""

    def __init__(self):
        super().__init__("tennis", "atp")


class UFC(League):
    """Provides access to UFC config files."""

    def __init__(self):
        super().__init__("fighting", "ufc")



from pysbr.queries.query import Query


class LeagueMarkets(Query):
    """Get the betting markets available for a particular league.

    Only market ids are returned by the query, without market names. For some leagues,
    this query returns many more market ids than the markets found on SBR.

    Args:
        league_id: SBR league id.
    """

    @Query.typecheck
    def __init__(self, league_id: int):
        super().__init__()
        self.name = "leagueMarkets"
        self.arg_str = self._get_args("league_id")
        self.args = {"lid": [league_id]}
        self.fields = self._get_fields("league_market")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._id_key = "market id"



from typing import Dict, List, Union
from collections import OrderedDict

import pysbr.utils as utils
from pysbr.config.config import Config


class Sportsbook(Config):
    """Provides access to sportsbook config file.

    In the config file there is information provided for the 27 sportsbooks found on
    SBR.

    Note that the id returned from lines-related queries is called 'sportsbook id' by
    this application, which is translated from 'paid', the name returned from SBR.
    There is another sportsbook id that is only used by the 'Sportsbooks' query, that
    is called 'system sportsbook id' by this application, which is translated from
    'sbid', the name returned from SBR. The system sportsbook id is not used by other
    parts of the application.

    Attributes:
        names (Dict[int, str]): Map sportsbook id to name. Used by Query.list() and
            Query.dataframe() in order to translate from id to name.
    """

    def __init__(self):
        super().__init__()

        self._sportsbooks = self._translate_dict(
            utils.load_yaml(utils.build_yaml_path("sportsbooks"))
        )
        self._sportsbook_ids = self._build_sportsbook_ids()

        self.names = {
            x["sportsbook id"]: x["name"] for x in self._sportsbooks["sportsbooks"]
        }

    def _build_sportsbook_ids(self) -> Dict[str, Dict[str, int]]:
        """Build sportsbook id search dictionary."""

        s = self._sportsbooks["sportsbooks"]
        sportsbooks = {}
        for k in ["name", "alias"]:
            sportsbooks[k] = {}
            for x in s:
                sportsbooks[k][x[k].lower()] = x["sportsbook id"]

        return sportsbooks

    def sportsbook_config(self) -> List[Dict[str, Union[str, int]]]:
        """Get sportsbook config list.

        Each list element is a dict representing a sportsbook, with 'name',
        n 'short name', 'sportsbook id' and 'system sportsbook id' as keys.
        """
        return self._sportsbooks

    def ids(self, terms: Union[List[Union[int, str]], int, str]) -> List[int]:
        """Take provided search terms and return list of matching sportsbook ids.

        If search term is string, search for matching sportsbook. If search term is
        int, assume that it is the ID, and insert it into the list to be returned.

        This method is provided as a convenience so that you don't need to
        remember sportsbook id numbers. Case is ignored for search terms.

        Example search terms:
            'pinnacle'
            'PINNACLE'
            'bodog'
            'bodog sportsbook'
            20

        Raises:
            TypeError:
                If a provided search term is not an int or str.
            ValueError:
                If a provided search term string cannot be matched with a sportsbook.
        """
        terms = utils.make_list(terms)
        ids = []
        for t in terms:
            if isinstance(t, int):
                ids.append(t)
            else:
                old_t = t
                try:
                    t = t.lower()
                except AttributeError:
                    raise TypeError("Search terms must be ints or strings.")
                try:
                    id = [v[t] for k, v in self._sportsbook_ids.items() if t in v][0]
                    ids.append(id)
                except IndexError:
                    raise ValueError(f"Could not find sportsbook {old_t}.")

        return list(OrderedDict.fromkeys(ids))



from pysbr.queries.query import Query


class Team(Query):
    """Get information about a team from its team id.

    The team's name, location, conference, division, league, and associated ids are
    among the values returned from the query.

    Args:
        team_id: SBR team id.
    """

    @Query.typecheck
    def __init__(self, team_id: int):
        super().__init__()
        self.name = "team"
        self.arg_str = self._get_args("team_id")
        self.args = {"tmid": team_id}
        self.fields = self._get_fields("team")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )



from pysbr.queries.query import Query


class SearchLeagues(Query):
    """Get up to 5 leagues matching a given search term.

    The search term can be part of a league's full name or abbreviation. Case is
    ignored. The max number of upcoming events returned is enforced by the server.

    The query response includes league name, region, and id, as well as sport id.

    Example search terms:
        'NFL'
        'National Football League'
        'National'
        'nfl'

    Args:
        search_term: String referring to the league of interest.
    """

    @Query.typecheck
    def __init__(self, search_term: str):
        super().__init__()
        self.name = "multipleSearch"
        self.arg_str = self._get_args("multiple_search")
        self.args = {"search_term": [search_term]}
        self.fields = self._get_fields("search_league")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["searchLeague"]
        self._id_key = "league id"



from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class Sportsbooks(Query):
    """Get information about a number of sportsbooks from their system ids.

    Note that the id returned from lines-related queries is called 'sportsbook id' by
    this application, which is translated from 'paid', the name returned from SBR.
    There is another sportsbook id that is only used by the 'Sportsbooks' query, that
    is called 'system sportsbook id' by this application, which is translated from
    'sbid', the name returned from SBR. The system sportsbook id is not used by other
    parts of the application.

    Sportsbook name, id (paid), and system id (sbid) are included in the response.

    Args:
        system_sportsbook_ids:  The system ids of the sportsbooks of interest.
    """

    @Query.typecheck
    def __init__(self, system_sportsbook_ids: Union[List[int], int]):
        super().__init__()
        system_sportsbook_ids = utils.make_list(system_sportsbook_ids)
        self.name = "sportsbooks"
        self.arg_str = self._get_args("sportsbooks")
        self.args = {"sbids": system_sportsbook_ids}
        self.fields = self._get_fields("sportsbooks")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._id_key = "sportsbook id"



from pysbr.queries.query import Query


class LeagueHierarchy(Query):
    """Get the league hierarchy for a particular league.

    The query returns a list of dicts, each representing a team in the league. Each
    dict contains the team id, and information about what conference and division the
    team belongs to.

    Many leagues return an empty list, for example, ATP and many soccer leagues.

    Args:
        league_id: SBR league id.
    """

    @Query.typecheck
    def __init__(self, league_id: int):
        super().__init__()
        self.name = "leagueHierarchy"
        self.arg_str = self._get_args("league_hierarchy")
        self.args = {"lids": [league_id]}
        self.fields = self._get_fields("league_hierarchy")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._id_key = "team id"



from typing import Optional, List, Union, Tuple, Dict
from datetime import datetime

from pysbr.queries.query import Query
import pysbr.utils as utils


class EventsByParticipants(Query):
    """Get events for a number of participants over a given date range.

    The query returns all events in the date range in which at least one of the
    participants in the participant id list competed in.

    This query makes two requests to the server; one to get all the event and
    participant ids over the date range, and then another to get the full event
    information for matching events.

    Because of the first query, either a league or sport id must be provided.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    Args:
        participant_ids: SBR participant id or list of participant ids. A participant
            id may refer to a team or an individual.
        start: Python datetime object representing the start date to search.
        end: Python datetime object representing the end date to search.
        league_id: SBR league id.
        sport_id: SBR sport id.
    """

    @Query.typecheck
    def __init__(
        self,
        participant_ids: Union[List[int], int],
        start: datetime,
        end: datetime,
        league_id: Optional[int] = None,
        sport_id: Optional[int] = None,
    ):
        if league_id is None and sport_id is None:
            raise ValueError("Either league_id or sport_id must not be None.")

        super().__init__()
        utils.make_list(participant_ids)
        event_ids = self._filter_events(
            participant_ids, start, end, league_id, sport_id
        )
        if not event_ids:
            return None

        self.name = "eventsV2"
        self.arg_str = self._get_args("event_ids")
        # TODO
        self.args = {"eids": event_ids}
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"

    def _league_args(
        self, start: datetime, end: datetime, league_id: int
    ) -> Tuple[str, Dict[str, Union[List[int], int]]]:
        """Get the arguments needed to query by league."""
        q_arg_str = self._get_args("date_range")
        q_args = {
            "lids": [league_id],
            "start": utils.datetime_to_timestamp(start),
            "end": utils.datetime_to_timestamp(end),
        }
        return q_arg_str, q_args

    def _sport_args(
        self, start: datetime, end: datetime, sport_id: int
    ) -> Tuple[str, Dict[str, Union[List[int], int]]]:
        """Get the arguments needed to query by sport."""
        q_arg_str = self._get_args("date_range_sport")
        q_args = {
            "spids": [sport_id],
            "start": utils.datetime_to_timestamp(start),
            "end": utils.datetime_to_timestamp(end),
        }
        return q_arg_str, q_args

    def _filter_events(
        self,
        participant_ids: List[int],
        start: datetime,
        end: datetime,
        league_id: int,
        sport_id: int,
    ) -> List[int]:
        """Make a query to get all events, and then filter out the relevant ones.

        The query gets all the events for a league or a sport over the given date range,
        and then a list of ids of events including at least one participant in the list
        is returned.
        """
        q_name = "eventsV2"
        q_arg_str, q_args = (
            self._league_args(start, end, league_id)
            if league_id is not None
            else self._sport_args(start, end, sport_id)
        )
        q_fields = self._get_fields("event_2")
        raw = self._build_and_execute_query(q_name, q_fields, q_arg_str, q_args)

        # filter out relevant events
        ids = []
        for e in raw["eventsV2"]["events"]:
            try:
                event_pids = [p["partid"] for p in e["participants"]]
                for pid in participant_ids:
                    if pid in event_pids:
                        ids.append(e["eid"])
            except KeyError:
                pass

        return ids



from pysbr.queries.query import Query


class SearchSports(Query):
    """Get up to 5 sports matching a given search term.

    Case is ignored. The max number of upcoming events returned is enforced by the
    server.

    The query response includes sport name and id.

    Args:
        search_term: String referring to the league of interest.
    """

    @Query.typecheck
    def __init__(self, search_term: str):
        super().__init__()
        self.name = "multipleSearch"
        self.arg_str = self._get_args("search_sport")
        self.args = {"search_term": [search_term]}
        self.fields = self._get_fields("search_sport")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["searchSport"]
        self._id_key = "sport id"



import pysbr.utils as utils
from typing import Dict


class Config:
    """Base config class to be subclassed by other config classes."""

    def __init__(self):
        self._translations = utils.load_yaml(utils.build_yaml_path("dictionary"))

    def _translate_dict(self, d: Dict) -> Dict:
        """Given a dict read from config file, make field names more readable.

        The field names in the config files are mostly abbreviations from SBR, which
        aren't nice to read, so translate all dictionary keys that are SBR abbrevations
        into English words.
        """

        def _recurse(el):
            """Recursive method used to iterate over all dict keys."""
            if isinstance(el, dict):
                # MUST cast to list to avoid RuntimeError because d.pop()
                for k in list(el.keys()):
                    try:
                        old_k = k
                        k = t[k]

                        el[k] = el.pop(old_k)
                    except KeyError:
                        pass
                    v = el[k]
                    if isinstance(v, dict) or isinstance(v, list):
                        _recurse(v)
            elif isinstance(el, list):
                for x in el:
                    _recurse(x)

        t = self._translations
        _recurse(d)
        return d

    def translations(self) -> Dict[str, str]:
        """Get the dict containing translations from abbrevations into words.

        The keys are abbreviations are from SBR, and the values are their translations
        into words.
        """
        return self._translations



from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class MarketsByMarketIds(Query):
    """Get information about a number of leagues from their league ids.

    Market name, description, and market type id are included in the response.

    Args:
        market_ids: SBR market id or list of market ids.
        sport_id: SBR sport id.
    """

    @Query.typecheck
    def __init__(self, market_ids: Union[List[int]], sport_id: int):
        super().__init__()
        market_ids = utils.make_list(market_ids)
        self.name = "marketTypesById"
        self.arg_str = self._get_args("market_ids")
        self.args = {"mtids": market_ids, "spids": [sport_id]}
        self.fields = self._get_fields("markets_by_id")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._id_key = "market id"



from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class LeaguesByLeagueIds(Query):
    """Get information about a number of leagues from their league ids.

    League name, abbreviation, and sport id are included in the response.

    Args:
        league_ids: SBR league id or list of league ids.
    """

    @Query.typecheck
    def __init__(self, league_ids: Union[List[int], int]):
        super().__init__()
        league_ids = utils.make_list(league_ids)
        self.name = "leagues"
        self.arg_str = self._get_args("league_ids")
        self.args = {"lids": league_ids}
        self.fields = self._get_fields("leagues")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )



from typing import List, Union

from pysbr.queries.lines import Lines
from pysbr.queries.query import Query
import pysbr.utils as utils


class OpeningLines(Lines):
    """Get the opening lines offered by a sportsbook for a number of events and markets.

    The date and time that the line was offered is also recorded. Both American and
    decimal odds are included.

    Args:
        event_ids: SBR event id or list of event ids.
        market_ids: SBR betting market id or list of market ids.
        sportsbook_id: SBR sportsbook id.
    """

    @Query.typecheck
    def __init__(
        self,
        event_ids: Union[List[int], int],
        market_ids: Union[List[int], int],
        sportsbook_id: int,
    ):
        super().__init__()
        event_ids = utils.make_list(event_ids)
        market_ids = utils.make_list(market_ids)
        self.name = "openingLines"
        self.arg_str = self._get_args("lines_2")
        self.args = {
            "eids": event_ids,
            "mtids": market_ids,
            "paid": sportsbook_id,
        }
        self.fields = None
        self._raw = self._build_and_execute_query(
            self.name, q_arg_str=self.arg_str, q_args=self.args
        )



from typing import List, Union

from pysbr.queries.lines import Lines
from pysbr.queries.query import Query
import pysbr.utils as utils


class CurrentLines(Lines):
    """Get the current lines offered by sportsbooks for a number of events and markets.

    For each sportsbook, event, participant and market combination, the most recent
    line offered by the book is in the response. Both American and decimal odds are
    included.

    Args:
        event_ids: SBR event id or list of event ids.
        market_ids: SBR betting market id or list of market ids.
        sportsbook_ids: SBR sportsbook id or list of sportsbook ids.
    """

    @Query.typecheck
    def __init__(
        self,
        event_ids: Union[List[int], int],
        market_ids: Union[List[int], int],
        sportsbook_ids: Union[List[int], int],
    ):
        super().__init__()
        event_ids = utils.make_list(event_ids)
        market_ids = utils.make_list(market_ids)
        sportsbook_ids = utils.make_list(sportsbook_ids)
        self.name = "currentLines"
        self.arg_str = self._get_args("lines_3")
        self.args = {
            "eids": event_ids,
            "mtids": market_ids,
            "paids": sportsbook_ids,
        }
        self.fields = None
        self._raw = self._build_and_execute_query(
            self.name, q_arg_str=self.arg_str, q_args=self.args
        )



from string import Template
import copy
import typing
from typing import Callable, Any, Dict, Optional, List, Union
from functools import wraps

from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
import pandas as pd

import pysbr.utils as utils
from pysbr.config.config import Config


class Query:
    """Base class for making queries on the SBR GraphQL endpoint.

    This class should not be directly instantiated; use the subclasses defined for each
    query.
    """

    def __init__(self):
        self._config = Config()

        self._raw = None
        self._subpath_keys = None
        self._sublist_keys = None
        self._id_key = None

        self._translated = None

        self._arguments = utils.load_yaml((utils.build_yaml_path("arguments")))
        self._fields = utils.load_yaml((utils.build_yaml_path("fields")))

        transport = RequestsHTTPTransport(
            url="https://www.sportsbookreview.com/ms-odds-v2/odds-v2-service"
        )
        self.client = Client(transport=transport, fetch_schema_from_transport=False)

    def typecheck(f: Callable) -> Callable:
        """Decorator for type checking arguments passed to subclass __init__ methods.

        The purpose of decorating the subclass __init__ methods is to avoid making
        invalid queries on the GraphQL endpoint.

        This method is only verified to work with List, Union, and primitive types.

        Raises:
            TypeError: If argument does not match expected type.
        """

        def recurse(a: Any, t: Dict) -> bool:
            """Method for recursively validating nested types."""
            t_origin = typing.get_origin(t)
            t_args = typing.get_args(t)

            if t_origin is None and len(t_args) == 0:
                if not isinstance(a, t):
                    return False

            elif isinstance(t_origin, type(list)):
                if not isinstance(a, list):
                    return False
                if len(t_args) > 0:
                    try:
                        for x in a:
                            for arg in t_args:
                                if not recurse(x, arg):
                                    return False
                    except TypeError:
                        return False
                else:
                    if not isinstance(a, t_origin):
                        return False

            elif isinstance(t_origin, typing._SpecialForm):
                for arg in t_args:
                    is_arg = recurse(a, arg)
                    if is_arg:
                        return True
                return False

            return True

        @wraps(f)
        def wrapper(*args: Any) -> Any:
            """Wrapper returned by the decorator, wrapping the function argument."""
            types = list(typing.get_type_hints(f).values())
            # first argument is self, ignore it
            for a, t in zip(args[1:], types):
                valid = recurse(a, t)
                if not valid:
                    raise TypeError(f"Expected {t}, got {a}")

            f(*args)

        return wrapper

    def _build_args(self, arg_str: str, args: Dict[str, Any]) -> Optional[str]:
        """Build the argument string that gets inserted into a query string.

        arg_str should be a string with substitution placeholders matching each arg in
        args.

        Raises:
            KeyError: If keys in args do not match placeholders in arg_str.
            ValueError: If arg_str is not a valid Template string.
        """
        if arg_str is not None and args is not None:
            return Template(arg_str).substitute(args)
        else:
            return None

    def _build_query_string(
        self, q_name: str, q_fields: Optional[str] = None, q_args: Optional[str] = None
    ) -> str:
        """Build up the GraphQL query string from given parameters.

        q_args should be the arguments to the query, with the values already filled in
        (i.e. query._build_args() should be called first).

        Returns:
            The completed query string ready to be executed.
        """
        return (
            Template(
                utils.str_format(
                    """
                query {
                    $q_name(
                        $q_args
                    ) $q_fields
                }
            """
                )
            ).substitute(
                {
                    "q_name": q_name,
                    "q_args": ""
                    if q_args is None
                    else utils.str_format(q_args, indent_=2, dedent_l1=True),
                    "q_fields": ""
                    if q_fields is None
                    else utils.str_format(q_fields, indent_=1, dedent_l1=True),
                }
            )
            # graphql query will not accept single quotes, but Template string by
            # default uses single quotes
            .replace("'", '"')
        )

    def _get_args(self, k: str) -> str:
        """Get the value of key k from the arguments dictionary.

        Raises:
            KeyError: If k is not a key in the arguments dict.
        """
        return self._arguments[k]

    def _get_fields(self, k: str) -> str:
        """Get the value of key k from the fields dictionary.

        Raises:
            KeyError: If k is not a key in the fields dict.
        """
        return self._fields[k]

    def _execute_query(self, q: str) -> Dict:
        """Execute the GraphQL query specified by the string q.

        Raises:
            gql.GraphQLSyntaxError: If the query string is structured improperly.
            gql.Exception: If the server raises an error during execution of the query.
        """
        return self.client.execute(gql(q))

    def _build_and_execute_query(
        self,
        q_name: str,
        q_fields: Optional[str] = None,
        q_arg_str: Optional[str] = None,
        q_args: Optional[Dict[str, Any]] = None,
    ) -> Dict:
        """Build query out of parameters, then execute it.

        Raises:
            gql.GraphQLSyntaxError: If the query string is structured improperly.
            gql.Exception: If the server raises an error during execution of the query.
            KeyError: If keys in q_args do not match placeholders in q_arg_str.
            ValueError: If q_arg_str is not a valid Template string.
        """
        q_string = self._build_query_string(
            q_name, q_fields, self._build_args(q_arg_str, q_args)
        )
        return self._execute_query(q_string)

    def _find_data(self):
        """Return a reference to to the relevant part of the query response.

        self._subpath_keys is a list of strings, which are a sequence of keys pointing
        to the data.
        """
        data = self._raw[self.name]
        if self._subpath_keys is not None:
            for k in self._subpath_keys:
                data = data[k]

        return data

    def _translate_dict(self, d: Dict) -> Dict:
        """Use translations from Config class to translate GraphQL response.

        This method is used by self.list() and self.dataframe() in order to translate
        field names from SBR into English words. Timestamps are converted into ISO
        strings.
        """

        def _recurse(el):
            """Recursive method used to iterate over all dict keys."""
            if isinstance(el, dict):
                # MUST cast to list to avoid RuntimeError because d.pop()
                for k in list(el.keys()):
                    try:
                        old_k = k
                        # raises KeyError if no translation available
                        k = t[k]

                        v = el.pop(old_k)
                        try:
                            # Sometimes values that should be integers are returned as
                            # strings. Need to make sure we're not truncating floats,
                            # however.
                            if isinstance(v, str):
                                v = int(v)
                        except ValueError:
                            pass
                        if k in ["datetime", "start datetime", "end datetime"]:
                            try:
                                # SearchEvents returns Zulu time iso string for value of
                                # 'date'.
                                v = utils.iso_str_to_timestamp(
                                    utils.iso_zulu_to_offset(v)
                                )
                            except (AttributeError, ValueError):
                                # AttributeError raised by iso_zulu_to_offset if v is a
                                # timestamp. ValueError raised by iso_str_to_timestamp
                                # if v is invalid isoformat string.
                                pass
                            try:
                                el[k] = utils.timestamp_to_iso_str(v)
                            except TypeError:
                                el[k] = v
                        else:
                            el[k] = v
                    except KeyError:
                        pass
                    v = el[k]
                    if isinstance(v, dict) or isinstance(v, list):
                        _recurse(v)
            elif isinstance(el, list):
                for x in el:
                    _recurse(x)

        t = self._config.translations()
        _recurse(d)
        return d

    def _copy_and_translate_data(self) -> List[Dict]:
        """Translate SBR fields in GraphQL response, and return a copy.

        This method is used by self.list() and self.dataframe(). self._translated
        caches the translated data.
        """
        if self._translated is None:
            data = copy.deepcopy(self._find_data())
            self._translated = self._translate_dict(data)
        return copy.deepcopy(self._translated)

    def arguments(self) -> Dict[str, str]:
        """Get the arguments dictionary, containing templates for all subqueries."""
        return self._arguments

    def fields(self) -> Dict[str, str]:
        """Get the fields dictionary, containing templates for all subqueries."""
        return self._fields

    def raw(self) -> Dict:
        """Get the raw GraphQL response, without any data processing."""
        return self._raw

    def ids(self) -> List[int]:
        """Get a list of ids from the query response.

        The type of id returned depends on the Query implementation. For example,
        calling this method on queries returning events will return a list of event ids.

        This method is useful for queries accepting lists of ids as arguments.

        Raises:
            NotImplementedError: If the Query object does not have a default return id
                type.
        """
        if self._id_key is None:
            raise NotImplementedError(
                f"{type(self).__name__} does not have a default return id type."
            )

        translated = self.list()
        ids = []
        for el in translated:
            ids.append(el[self._id_key])

        return list(set(ids))

    def list(self) -> List[Dict[str, Union[str, List, Dict]]]:
        """Get a list of translated elements returned from the query.

        Each element in the response is a dict with fields requested in the query and
        the element's values for those fields. The fields are translated.
        """
        data = self._copy_and_translate_data()
        # Some queries return dictionaries. Enforce this method returning a list.
        if isinstance(data, dict):
            data = [data]
        return data

    def dataframe(self) -> pd.DataFrame:
        """Get a dataframe of elements returned from the query.

        Each element in the response is a dict with fields requested in the query and
        the element's values for those fields. The fields are translated, and if the
        elements are nested, it is attempted to flatten them using self._sublist_keys,
        which is a list of keys expected to be in each element that have values that
        are lists.

        If the elements cannot be flattened, a dataframe of the nested elements is
        returned.
        """
        data = self._copy_and_translate_data()

        # Using sublist_keys instead of recursive method because there is a possibility
        # of overwriting keys without realizing it if using recursive method.
        # The idea is that pd.json_normalize() doesn't work on sublists.
        if self._sublist_keys is not None:
            for k in self._sublist_keys:
                for el in data:
                    try:
                        for i, subel in enumerate(el.get(k)):
                            new_key = f"{k}.{i+1}"
                            el[new_key] = subel
                        el.pop(k)
                    except TypeError:
                        pass
                    except AttributeError:
                        pass
        try:
            return pd.json_normalize(data)
        except AttributeError:
            return pd.DataFrame(data)



from pysbr.queries.query import Query


class EventMarkets(Query):
    """Get the event markets available on a particular event.

    The response contains only the market ids, not their names.

    Args:
        event_id: SBR event id.
    """

    @Query.typecheck
    def __init__(self, event_id: int):
        super().__init__()
        self.name = "eventMarkets"
        self.arg_str = self._get_args("event_id")
        self.args = {"eid": event_id}
        self.fields = self._get_fields("market_types")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["mtids"]



import copy
from typing import List, Dict, Union, Tuple

import pandas as pd

from pysbr.queries.query import Query
from pysbr.config.sport import (
    NFL,
    NCAAF,
    ATP,
    Bundesliga,
    EPL,
    LaLiga,
    MLB,
    NBA,
    NCAAB,
    NHL,
    UCL,
    UEFANationsLeague,
    UFC,
)
from pysbr.config.sport import (
    Football,
    Basketball,
    Baseball,
    Hockey,
    Soccer,
    Tennis,
    Fighting,
)
from pysbr.config.sport import Sport
from pysbr.config.sportsbook import Sportsbook


class Lines(Query):
    """Implements methods particular to queries about betting lines.

    This class should not be directly instantiated; use the subclasses defined for each
    lines-related query.
    """

    def __init__(self):
        self._events = None
        self._event_descriptions = {}
        self._event_leagues = {}
        self._event_sports = {}
        self._event_scores = {}
        self._event_statuses = {}
        # these are the participant ids for Over/Under lines for all sports I checked
        self._participants = {15143: "over", 15144: "under"}
        self._leagues = {
            16: NFL,
            6: NCAAF,
            23: ATP,
            11: Bundesliga,
            2: EPL,
            17: LaLiga,
            5: NBA,
            3: MLB,
            14: NCAAB,
            7: NHL,
            8: UCL,
            1911: UEFANationsLeague,
            26: UFC,
        }
        self._leagues_init = {}

        self._sports = {
            4: Football,
            5: Basketball,
            3: Baseball,
            6: Hockey,
            1: Soccer,
            8: Tennis,
            9: Fighting,
        }
        self._sports_init = {}

        self._sportsbooks = None

        self._with_ids_translated = None

        super().__init__()

    def _clean_lines(self, data: List[Dict]) -> List[Dict]:
        """Remove unneeded keys from the query response.

        This is necessary for lines-related queries because they don't accept any
        fields, so some unneeded fields are returned.
        """
        to_remove = [
            "boid",
            "lineid",
            "sequence",
            "dp",
            "bs",
            "iof",
            "sbid",
            "sid",
            "fpd",
            "fpn",
            "sort",
        ]
        for term in to_remove:
            for line in data:
                try:
                    # ConsensusHistory has 'line' as a key, instead of being the
                    # top-level dictionary.
                    line = line["line"]
                except KeyError:
                    pass
                try:
                    # The earliest (time-wise) lines in consensusHistory may not have
                    # 'line' as a key.
                    line.pop(term, None)
                except AttributeError:
                    pass
        return data

    def _init_config(self, data: List[Dict]) -> None:
        """Initialize private instance variables.

        There is a lot of data that needs to be initialized in order to translate from
        ids to information. This method does that.

        self._leagues and self._sports map league and sport ids to their configuration
        classes. Other private instance variables map event and participant ids to their
        translations.
        """
        if self._sportsbooks is None:
            self._sportsbooks = Sportsbook().names

        league_ids = [e.get("league id") for e in self._events.list()]
        for id in set(league_ids):
            try:
                self._leagues_init[id] = self._leagues[id]()
            except KeyError:
                pass

        sport_ids = [e.get("sport id") for e in self._events.list()]
        for id in set(sport_ids):
            try:
                self._sports_init[id] = self._sports[id]()
            except KeyError:
                pass

        for e in self._events.list():
            self._event_descriptions[e.get("event id")] = e.get("description")
            self._event_leagues[e.get("event id")] = e.get("league id")
            self._event_sports[e.get("event id")] = e.get("sport id")
            self._event_scores[e.get("event id")] = e.get("scores")
            self._event_statuses[e.get("event id")] = e.get("event status")
            try:
                for p in e["participants"]:
                    participant_id = p.get("participant id")
                    source = p.get("source")
                    if "last name" in source and source.get("last name") is not None:
                        self._participants[participant_id] = source.get("last name")
                    elif (
                        "abbreviation" in source
                        and source.get("abbreviation") is not None
                    ):
                        self._participants[participant_id] = source.get("abbreviation")
                    elif "name" in source and source.get("name") is not None:
                        self._participants[participant_id] = source.get("name")

            except KeyError:
                # Should not reach here; e['participants'] should be empty list at
                # least.
                pass

    def _get_config(self, line: List[Dict]) -> Sport:
        """Get league or sport config class.

        If neither the league nor the sport has a configuration class, then the market
        name and bet result will not appear in the Query.list() or Query.dataframe().
        """
        try:
            # League needs to be before sport because it may override some attributes.
            # E.g. NCAAB market periods.
            return self._leagues_init[self._event_leagues.get(line.get("event id"))]
        except KeyError:
            try:
                return self._sports_init[self._event_sports.get(line.get("event id"))]
            except KeyError:
                # neither league nor sport has a config file
                return None

    def _resolve_market(self, line: List[Dict]) -> str:
        """Attempt to get the name of the market from its id."""
        try:
            return self._get_config(line).market_names.get(line.get("market id"))
        except AttributeError:
            return None

    def _tally_points(
        self,
        line: List[Dict],
        period_scores: List[Dict[str, int]],
        market_range: List[int],
    ) -> Tuple[int, int]:
        """Sum up the points scored over the range of interest, according to the market
        in question.

        If the market is a total, the return is of the form (total, []). Otherwise, the
        return looks like (points_scored_by_team, points_scored_by_other_team).
        """
        scores = copy.deepcopy(period_scores)
        # TODO: check that len scores == market range (current and future events)
        if market_range is not None:
            scores = [s for s in period_scores if s.get("period") in market_range]

        participant_id = line.get("participant id")
        o_scores = []
        if participant_id not in [15143, 15144]:
            scores = [
                s for s in period_scores if s.get("participant id") == participant_id
            ]
            o_scores = [
                s for s in period_scores if s.get("participant id") != participant_id
            ]

        try:
            return sum([s.get("points scored") for s in scores]), sum(
                [s.get("points scored") for s in o_scores]
            )
        except TypeError:
            return None, None

    def _evaluate_bet(self, line: List[Dict], points, o_points):
        """Evaluate whether the bet won or lost.

        Given a line, market, and point totals, determine whether the bet won or lost,
        returning True if win, and False if lose.
        """
        try:
            market_type = self._get_config(line).market_types.get(line.get("market id"))
        except AttributeError:
            return None

        participant_id = line.get("participant id")
        spread_or_total = line["spread / total"]
        if market_type == "total":
            # participant ids for o/u
            over = 15143
            under = 15144
            if (
                participant_id == over
                and points > spread_or_total
                or participant_id == under
                and points < spread_or_total
            ):
                return True
            else:
                return False
        else:
            try:
                if points + spread_or_total > o_points:
                    return True
                else:
                    return False
            except ValueError:
                return None

    def _resolve_bet(self, line: List[Dict]) -> Tuple[str, float]:
        """Given a line, determine what the result of the bet was.

        Utilizes event scores list which is part of an event query response. This
        method calculates the point totals for the team of the line in question vs. the
        other team, over the periods / quarters that the bet is evaluated, in order to
        determine the result of the bet.

        Returns a tuple where the first value is 'W' or 'L', and the second value is
        the amount of profit the bet would return on a $100 bet. If some step fails,
        (None, None) is returned and the bet result / profit are not included for that
        particular line. Errors are not raised.
        """
        event_status = self._event_statuses.get(line.get("event id"))
        scores = self._event_scores.get(line.get("event id"))
        # Event query didn't have scores, or the game hasn't been played yet (query
        # returns empty list), or event is in progress.
        if not scores or event_status != "complete":
            return None, None, None

        try:
            market_periods = self._get_config(line).market_periods.get(
                line.get("market id")
            )
        except AttributeError:
            return

        try:
            market_range = list(range(market_periods[0], market_periods[-1]))
        except TypeError:
            market_range = None

        points, o_points = self._tally_points(line, scores, market_range)
        if points is None or o_points is None:
            return None, None, None

        is_win = self._evaluate_bet(line, points, o_points)
        if is_win is None:
            return None, None, None

        try:
            profit = (
                round((line.get("decimal odds") - 1) * 100, 2) if is_win else -100.0
            )
        except ValueError:
            return None, None, None

        return ("W" if is_win else "L", profit, points)

    def _translate_ids(self, data: List[Dict]) -> List[Dict]:
        """Add new entries to each element in the list for the element's id fields.

        The response for lines-related queries has many ids without associated
        information, making it hard to remember which line is related to which event.
        This method adds that related information to each element in the list as long
        as a list of events has been passed in when calling self.list() or
        self.dataframe().

        If a list of events (that the lines come from) is not passed to self.list() or
        self.dataframe(), this method has no effect. Otherwise, it adds the following
        to each element:
            event description
            betting market name
            sportsbook name
            participant information
            bet result information (whether it won or lost; only for completed events)

        self._with_ids_translated caches the returned list.
        """
        if self._events is None:
            return data

        if self._with_ids_translated is not None:
            return self._with_ids_translated

        self._init_config(data)

        for line in data:
            line["event"] = self._event_descriptions.get(line.get("event id"))
            market = self._resolve_market(line)
            if market is not None:
                line["market"] = market
            result, profit, points = self._resolve_bet(line)
            if result is not None:
                line["result"] = result
            if profit is not None:
                line["profit"] = profit
            if points is not None:
                line["participant score"] = points
            line["sportsbook"] = self._sportsbooks.get(line.get("sportsbook id"))
            line["participant"] = self._participants.get(line.get("participant id"))

        self._with_ids_translated = data
        return data

    def _copy_and_translate_data(self) -> List[Dict]:
        """Translate SBR fields in GraphQL response, and return a copy.

        This method is used by self.list() and self.dataframe(). Overrides Query.
        _copy_and_translate_data() in order to add steps for cleaning the response and
        translating the ids in the response.
        """
        data = copy.deepcopy(self._find_data())
        self._clean_lines(data)
        self._translate_dict(data)
        return self._translate_ids(data)

    def list(self, events=None) -> List[Dict[str, Union[str, List, Dict]]]:
        """Get a list of translated elements returned from the query.

        If a list of events the lines are for is passed in, extra information about
        each line will be added to each element of the returned list, including event
        description, participant information, and betting market name.
        """
        self._events = events
        return super().list()

    def dataframe(self, events=None) -> pd.DataFrame:
        """Get a dataframe of elements returned from the query.

        If a list of events the lines are for is passed in, extra information about
        each line will be added to each row of the returned dataframe, including event
        description, participant information, and betting market name.
        """
        self._events = events
        return super().dataframe()



from typing import List, Union

from pysbr.queries.lines import Lines
from pysbr.queries.query import Query
import pysbr.utils as utils


class LineHistory(Lines):
    """Get the line history of an event and market for a particular sportsbook.

    The complete history of line movement for an event, market, and sportsbook is
    returned by the query. Both American and decimal odds are included.

    The query to the server does not work without at least one participant id of a
    participant in the event. If you only include 1 participant id, the query will only
    return the history for the side of the bet relevant to that participant id.

    Args:
        event_id: SBR event id.
        market_id: SBR betting market id.
        sportsbook_id: SBR sportsbook id.
        participant_ids: List of participant ids of participants in the event, or just
        one participant id.
    """

    @Query.typecheck
    def __init__(
        self,
        event_id: int,
        market_id: int,
        sportsbook_id: int,
        participant_ids: Union[List[int], int],
    ):
        # only need 1 participant id, it's dumb
        super().__init__()
        utils.make_list(participant_ids)
        self.name = "lineHistory"
        self.arg_str = self._get_args("line_history")
        self.args = {
            "eid": event_id,
            "mtid": market_id,
            "paid": sportsbook_id,
            # partids is required by the lineHistory query
            "partids": participant_ids,
        }
        self.fields = self._get_fields("line_history")

        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

    def _find_data(self):
        lines = self._raw[self.name]
        cleaned_lines = []
        for el in lines:
            cleaned_lines.append(el["lines"][0])

        return cleaned_lines



import json
import copy

from pysbr.queries.query import Query


# only returns five first results, and only for upcoming events
# cannot search by location, league etc. Only by event participants
class SearchEvents(Query):
    """Get up to 5 upcoming events matching a given search term.

    The search term should be a string referring to a team or participant's name. Case
    is ignored. The max number of upcoming events returned is enforced by the server.

    This query response has a different structure than other event queries, but still
    includes information about date and time, participants, description, and associated
    ids.

    Example search terms:
        'Seattle' (this will return information about all Seattle-based teams)
        'Seahawks'
        'Federer'
        'federer'
        'roger federer'
        'tiz the law' (horse racing)

    Args:
        search_term: String referring to team or participant of interest.

    """

    @Query.typecheck
    def __init__(self, search_term: str):
        super().__init__()
        self.name = "searchEvent"
        self.arg_str = self._get_args("search_event")
        self.args = {"search_term": [search_term]}
        self.fields = self._get_fields("search_event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._sublist_keys = ["participants"]
        self._id_key = "event id"

    def _string_to_json(self, data, key):
        for el in data:
            try:
                el[key] = json.loads(el[key])
            except TypeError:
                pass

    def _copy_and_translate_data(self):
        data = copy.deepcopy(self._find_data())
        self._string_to_json(data, "eventParticipants")
        return self._translate_dict(data)



from datetime import datetime
from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class EventsByDate(Query):
    """Get events for selected leagues on a certain date.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    Args:
        league_ids: SBR league id or list of league ids.
        dt: Python datetime object representing the date of interest. Events searched
            are those in the range [dt, dt + 24 hours].
    """

    @Query.typecheck
    def __init__(self, league_ids: Union[List[int], int], dt: datetime):
        super().__init__()
        league_ids = utils.make_list(league_ids)
        self.name = "eventsByDateNew"
        self.arg_str = self._get_args("date")
        self.args = {"lids": league_ids, "timestamp": utils.datetime_to_timestamp(dt)}
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"



from pysbr.queries.query import Query


class EventGroupsByLeague(Query):
    """Get the event groups available on a particular league.

    Event groups refer to things like Week 1 of the NFL season, or the US Open, so you
    can use this in conjunction with EventsByEventGroups to get information about a
    group of events. Event group id, name, start date and end date are included in the
    query response.

    Args:
        league_id: SBR league id.
    """

    @Query.typecheck
    def __init__(self, league_id: int):
        super().__init__()
        self.name = "eventGroupsByLeague"
        self.arg_str = self._get_args("league_id")
        self.args = {"lid": league_id}
        self.fields = self._get_fields("event_group")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._id_key = "event group id"



from pysbr.queries.query import Query


class EventsByEventGroup(Query):
    """Get events for a particular league and event group.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    season_id is a required argument because event group ids are not unique to a season.
    market_id is a required argument to the query object on the SBR server.

    Args:
        league_id: SBR league id.
        event_group_id: SBR event group id.
        season_id: SBR season id. The easiest way to find a season id is by making an
            event query about an event that belongs to the season in question, and
            getting the season id from that event.
        market_id: SBR market id. You can use a league's config class to get the
            default market id for a given sport.
    """

    @Query.typecheck
    def __init__(
        self, league_id: int, event_group_id: int, season_id: int, market_id: int
    ):
        super().__init__()
        self.name = "eventsByEventGroupV2"
        self.arg_str = self._get_args("event_group")
        self.args = {
            # breaks without league id
            "lid": league_id,
            # returns multiple seasons if not provided
            "seid": season_id,
            "egid": event_group_id,
            # mtid is a required argument for EventsByEventGroupV2
            "mtid": market_id,
        }
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"



from typing import List, Union

from pysbr.queries.lines import Lines
from pysbr.queries.query import Query
import pysbr.utils as utils


class BestLines(Lines):
    """Get the best lines offered by any sportsbook for a number of events and markets.

    For each event, participant and market combination, the best line offered by any of
    sportsbooks tracked by SBR is in the response. The date and time that the line was
    offered is also recorded. Both American and decimal odds are included.

    Sometimes the best line returned is from a sportsbook that is not active on SBR. In
    this case it is not easy to figure out which book it came from.

    Args:
        event_ids: SBR event id or list of event ids.
        market_ids: SBR betting market id or list of market ids.
    """

    @Query.typecheck
    def __init__(
        self, event_ids: Union[List[int], int], market_ids: Union[List[int], int]
    ):
        super().__init__()
        event_ids = utils.make_list(event_ids)
        market_ids = utils.make_list(market_ids)
        self.name = "bestLines"
        self.arg_str = self._get_args("lines")
        self.args = {"eids": event_ids, "mtids": market_ids}
        self.fields = None
        self._raw = self._build_and_execute_query(
            self.name, q_arg_str=self.arg_str, q_args=self.args
        )



from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class EventsByParticipantsRecent(Query):
    """Get the 5 most recent events for a number of participants.

    5 is an arbitrary number enforced by the query on the server side.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    Args:
        participant_ids: SBR participant id or list of participant ids.
    """

    @Query.typecheck
    def __init__(self, participant_ids: Union[List[int]]):
        # this query only gets the 5 most recent events for a participant.
        super().__init__()
        utils.make_list(participant_ids)
        self.name = "eventsInfoByParticipant"
        self.arg_str = self._get_args("participants")
        self.args = {"partids": participant_ids}
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"

    def _find_data(self):
        """Return a reference to to the relevant part of the query response.

        Overrides Query._find_data() because the structure of the returned data is
        different than most other queries.
        """
        data = self._raw[self.name]
        cleaned_data = []
        for participant in data:
            cleaned_data.extend(participant["events"])

        return cleaned_data



from pysbr.queries.bestlines import BestLines
from pysbr.queries.consensusHistory import ConsensusHistory
from pysbr.queries.currentlines import CurrentLines
from pysbr.queries.eventgroupsbyleague import EventGroupsByLeague
from pysbr.queries.eventmarkets import EventMarkets
from pysbr.queries.eventsbydate import EventsByDate
from pysbr.queries.eventsbydaterange import EventsByDateRange
from pysbr.queries.eventsbyeventids import EventsByEventIds
from pysbr.queries.eventsbymatchup import EventsByMatchup
from pysbr.queries.eventsbyparticipants import EventsByParticipants
from pysbr.queries.eventsbyparticipantsrecent import EventsByParticipantsRecent
from pysbr.queries.leaguehierarchy import LeagueHierarchy
from pysbr.queries.leaguemarkets import LeagueMarkets
from pysbr.queries.leaguesbyleagueids import LeaguesByLeagueIds
from pysbr.queries.linehistory import LineHistory
from pysbr.queries.marketsbymarketids import MarketsByMarketIds
from pysbr.queries.openinglines import OpeningLines
from pysbr.queries.searchevents import SearchEvents
from pysbr.queries.searchleagues import SearchLeagues
from pysbr.queries.searchsports import SearchSports
from pysbr.queries.sportsbooks import Sportsbooks
from pysbr.queries.team import Team

from pysbr.config.sport import (
    Football,
    Basketball,
    Baseball,
    Hockey,
    Soccer,
    Tennis,
    Fighting,
)
from pysbr.config.sport import (
    NFL,
    NCAAF,
    MLB,
    NBA,
    NCAAB,
    NHL,
    EPL,
    UCL,
    LaLiga,
    Bundesliga,
    UEFANationsLeague,
    ATP,
    UFC,
)
from pysbr.config.sportsbook import Sportsbook



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="python-sbr",
    version="0.1.1",
    description="Access the SportsbookReview GraphQL endpoint.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/JeMorriso/PySBR",
    author="Jeremy Morrison",
    author_email="jeremy.morrison36@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.8",
    packages=find_packages(exclude=["tests", "tests.*"]),
    keywords="""
        sportsbookreview sportsbook review betting api gambling sports graphql odds
        lines moneyline
    """,
    include_package_data=True,
    install_requires=["gql", "pandas", "pytz", "pyyaml"],
)



from typing import List, Union

from pysbr.queries.query import Query
import pysbr.utils as utils


class EventsByEventIds(Query):
    """Get events from a list of event ids.

    All event queries return information about matching events including date and time,
    location, participants, and associated ids.

    Args:
        event_ids: SBR event id or list of event ids.
    """

    @Query.typecheck
    def __init__(self, event_ids: Union[List[int], int]):
        super().__init__()
        utils.make_list(event_ids)
        self.name = "eventsV2"
        self.arg_str = self._get_args("event_ids")
        self.args = {"eids": event_ids}
        self.fields = self._get_fields("event")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )

        self._subpath_keys = ["events"]
        self._sublist_keys = ["participants", "scores"]
        self._id_key = "event id"



from typing import List, Union

from pysbr.queries.lines import Lines
from pysbr.queries.query import Query
import pysbr.utils as utils


class ConsensusHistory(Lines):
    """Get the consensus amongst SBR members for selected markets on an event.

    For events in some leagues and markets, SBR members can simulate a wager on the
    current line. The query response includes the number of wagers made on a line at a
    given price, and percentage of wagers made on each side of a bet. For certain
    markets volume' and 'total volume' are recorded where 'volume' refers to the average
    wager size, and 'total volume' is just volume multiplied by the # of wagers. For
    certain markets the number of whale bets made on a line is available.

    Note that the line available for members to make a wager is always from 5Dimes.
    The markets available for consensus history can be gotten from a league's config
    class, if it exists. In general, the 3 available markets are the moneyline, point
    spread, and total (all full-time).

    Args:
        event_id: SBR event id.
        market_ids: SBR betting market ids.
    """

    @Query.typecheck
    def __init__(self, event_id: int, market_ids: Union[List[int], int]):
        super().__init__()
        market_ids = utils.make_list(market_ids)
        self.name = "consensusHistory"
        self.arg_str = self._get_args("consensus_history")
        self.args = {"eid": event_id, "mtids": market_ids}
        self.fields = self._get_fields("consensus_history")
        self._raw = self._build_and_execute_query(
            self.name, self.fields, self.arg_str, self.args
        )



import json
from inspect import cleandoc
from textwrap import indent
from datetime import datetime
from pathlib import Path
import pathlib
from typing import Dict, Any, List

from pytz import timezone, utc
import yaml


def dump_json(d, path: str):
    """Write json to file specified by path.

    Raises:
        FileNotFoundError
    """
    with open(path, "w") as f:
        json.dump(d, f, indent=2)


def load_json(path) -> Dict:
    """Read json from file specified by path.

    Raises:
        FileNotFoundError
    """
    with open(path) as f:
        return json.load(f)


def load_yaml(path: str) -> Dict:
    """Read yaml from file specified by path.

    Raises:
        FileNotFoundError
    """
    with open(path) as f:
        return yaml.full_load(f)


def dump_yaml(d, path):
    """Write yaml to file specified by path.

    Raises:
        FileNotFoundError
    """
    with open(path, "w") as f:
        yaml.dump(d, f)


def get_project_root() -> pathlib.PosixPath:
    """Get pathlib representation of project's root folder."""
    return Path(__file__).parent.parent


def build_yaml_path(fname: str, subpath: str = "pysbr/config") -> pathlib.PosixPath:
    """Build an absolute path to the .yaml file specified by fname and subpath."""
    return get_project_root().joinpath(f"{subpath}/{fname}.yaml")


def str_format(
    s: str, indent_: int = 0, dedent_l1: bool = False, squish: bool = False
) -> str:
    """Format and return string s.

    Args:
        indent_: The number of tabs to indent each line of the string with.
        dedent_l1: Dedent line 1. This is added as an argument because if inserting a
            multiline string into a template string, only the first line gets the
            indentation from the template string.
        squish: If true, compress the string into one line.
    """
    tab = "    "
    s = indent(cleandoc(s), tab * indent_)
    if dedent_l1:
        lines = s.split("\n")
        lines[0] = lines[0].lstrip()
        s = "\n".join(lines)
    return s if not squish else s.replace("\n", " ")


def make_list(item: Any) -> List:
    """Return item inside a list, if it isn't already a list."""

    if not isinstance(item, list):
        return [item]
    else:
        return item


def datetime_to_timestamp_aware(dt: datetime, tz: str = None) -> int:
    """**DEPRECATED** Convert datetime object to Unix timestamp, in milliseconds.

    tz refers to the timezone to use (see pytz documentation for list of
            available timezones). If None, uses system timezone.
    """
    if dt.tzinfo is None:
        dt = timezone(tz).localize(dt) if tz is not None else dt.astimezone()
    return int(dt.timestamp() * 1000)


def timestamp_to_datetime_aware(ts: int, tz: str = None) -> datetime:
    """**DEPRECATED** Convert Unix timestamp ts to timezone aware datetime object.

    tz refers to the timezone to use (see pytz documentation for list of
            available timezones). If None, uses system timezone.
    """
    utc_dt = utc.localize(datetime.utcfromtimestamp(ts / 1000))
    return utc_dt.astimezone(timezone(tz) if tz is not None else None)


def datetime_to_timestamp(dt: datetime) -> int:
    """Convert datetime to Unix timestamp in milliseconds"""
    return int(dt.timestamp() * 1000)


def timestamp_to_datetime(ts: int) -> datetime:
    """Convert Unix timestamp (in milliseconds) to datetime.

    The returned datetime object is aware, and uses the system timezone.
    """
    return datetime.fromtimestamp(ts / 1000).astimezone()


def timestamp_to_iso_str(ts: int) -> str:
    """Convert Unix timestamp (in milliseconds) to ISO string."""

    return timestamp_to_datetime(ts).replace(microsecond=0).isoformat()


def iso_str_to_timestamp(iso_str: str) -> int:
    """Convert iso str to Unix timestamp in milliseconds."""
    return datetime.fromisoformat(iso_str).timestamp() * 1000


def iso_zulu_to_offset(iso_str: str) -> str:
    """Convert ISO string in Zulu time to UTC offset.

    Does not check the contents of the ISO string - assumes it is a correctly formatted
    ISO string in Zulu time.
    """
    return iso_str.replace("Z", "+00:00")
